from pydantic import BaseModel, Field, RootModel, ValidationError, field_validator
from typing import Any, Dict, List, Optional, Set

class TypeSpecificConfig(RootModel):
    pass

class BlockConfig(BaseModel):
    type: str
    version: str
    typeSpecificConfig: TypeSpecificConfig
    mode: Optional[str] = None

class Trigger(BaseModel):
    name: str
    event: str
    filter: str

class Output(BaseModel):
    name: str
    destination: str
    function: str
    parameters: Dict[str, str]

class Credential(BaseModel):
    type: str 
    value: str

class Stage(BaseModel):
    id: str = Field(..., max_length=255)
    type: str
    blockSpec: str
    blockConfig: BlockConfig
    credentials: Optional[Dict[str, Credential]] = {}
    triggers: List[Trigger] = []
    outputs: List[Output]
    memory: Optional[Dict[str, str]] = {}

class Pipeline(BaseModel):
    apiVersion: str = Field(..., alias='apiVersion')
    type: str
    metadata: Dict[str, Any]
    stages: List[Stage]
    @field_validator('metadata')
    def metadata_must_contain_name_and_description(cls, v):
        if 'name' not in v:
            raise ValueError("The 'metadata' field must contain the 'name' field.")
        if 'description' not in v:
            raise ValueError("The 'metadata' field must contain the 'description' field.")
        return v



from typing import Optional, TypedDict, List
from datetime import datetime
from enum import Enum
from pydantic import BaseModel

class ThreadStatus(Enum):
    UNTHREADED = 1
    THREAD_PARENT = 2
    THREAD_CHILD = 3

class InstantMessageReaction(BaseModel):
    name: str
    count: int
    users: List[str]
    parent: float

class EditedStatus(BaseModel):
    user: str
    ts: datetime

class InstantMessageFile(BaseModel):
    user: str # The user ID belonging to the user that incited this action.
    team: str
    channel: str
    internal_tenant: str

    file_name: str
    file_size: int
    ts: datetime
    file_url: str
    platform: str

class InstantMessage(BaseModel):
    platform: str # slack, teams, etc
    event_id: str # specific to the platform

    user: str # The user ID belonging to the user that incited this action.
    user_friendly_name: str # User's friendly name
    team: str
    channel: str
    channel_friendly_name: str

    text: str
    ts: str
    attachments: Optional[List[InstantMessageFile]] = None
    thread_ts: Optional[float] = None
    thread_status: ThreadStatus = ThreadStatus.UNTHREADED
    edited: Optional[EditedStatus] = None
    hidden: bool = False
    deleted: Optional[datetime] = None
    reactions: Optional[List[InstantMessageReaction]] = None



class MessagingInterface:
    """Defines the standard operations required for messaging platforms."""

    def parse_raw_message(self, raw_message: dict) -> InstantMessage:
        """Parses the incoming raw message to a structured internal representation."""
        raise NotImplementedError

    def parse_raw_react(self, raw_react: dict) -> InstantMessage:
        """Parses the incoming raw react to a structured internal representation."""
        raise NotImplementedError

    def parse_file_shared(self, raw_file_event: dict) -> InstantMessageFile:
        raise NotImplementedError

    def send_message(self, message: InstantMessage):
        """Sends a message using the platform's messaging API."""
        raise NotImplementedError

    def send_reaction(self, message_id: str, reaction_type: str):
        """Sends a reaction to a message based on its ID and the reaction type."""
        raise NotImplementedError

    def delete_message(self, message_id: str):
        """Deletes a message based on its ID."""
        raise NotImplementedError

    def edit_message(self, message_id: str, new_text: str):
        """Edits an existing message."""
        raise NotImplementedError


def cyclic(g: dict):
    """Return True if the directed graph g has a cycle.
    g must be represented as a dictionary mapping vertices to
    iterables of neighbouring vertices. For example:

    >>> cyclic({1: (2,), 2: (3,), 3: (1,)})
    True
    >>> cyclic({1: (2,), 2: (3,), 3: (4,)})
    False
    """
    path = set()
    visited = set()

    def visit(vertex):
        if vertex in visited:
            return False
        visited.add(vertex)
        path.add(vertex)
        for neighbour in g.get(vertex, ()):
            if neighbour in path or visit(neighbour):
                return True
        path.remove(vertex)
        return False

    return any(visit(v) for v in g)


import json
from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers.web import JsonLexer
def pretty_print_json(data: dict) -> None:
    raw_json = json.dumps(data, indent=4)
    colorful = highlight(
        raw_json,
        lexer=JsonLexer(),
        formatter=TerminalTrueColorFormatter(),
    )
    print(colorful)


from pydantic import BaseModel, Field, RootModel, ValidationError, field_validator
from typing import Any, Dict, List, Optional, Set

class ProfileCredentials(BaseModel):
    tenant: int
    token: str
    id_token: str


from pydantic import BaseModel, Field, RootModel, ValidationError, validator
from typing import Any, Dict, List, Optional, Set, TypeVar, Generic, Type, Callable, Literal, Union
from ..connector import BaseExecutionOutput
import yaml
import importlib
import functools
import inspect
from types import NoneType


class EventDefinition(BaseModel):
    event: str
    description: str


class FunctionParams(BaseModel):
    name: str
    require: bool
    description: str
    type: str

class FunctionReturn(BaseModel):
    name: str
    description: str
    type: str

class FunctionDefinition(BaseModel):
    name: str
    description: str
    parameters: List[FunctionParams]
    returns: Union[str, FunctionReturn]

class ActionsDefinition(BaseModel):
    triggers: Optional[List[EventDefinition]]
    functions: Optional[List[FunctionDefinition]]

class CredentialDefinition(BaseModel):
    handler: str
    required: bool = False

class BlockConfigDefinition(BaseModel):
    apiVersion: str = Field(..., alias='apiVersion')
    type: str
    metadata: Dict[str, Any]
    handler: str
    modes: List[str]
    credentials: Optional[Dict[str, CredentialDefinition]] = {}
    actions: ActionsDefinition
    pool: Optional[Literal['cpu-short', 'cpu-long', 'cpu-nointerrupt', 'gpu']] = 'cpu-short'
    timeout: Optional[int] = 90
                 
def read_block_definition(yaml_path: str) -> BlockConfigDefinition:
    with open(yaml_path, 'r') as file:
        data = yaml.safe_load(file)
    return BlockConfigDefinition(**data)

def register_trigger_handler(event_name: str) -> Callable:
    """Register a given function as a handler for a event trigger named `event_name`."""
    def decorator(func: Callable[..., BaseExecutionOutput]) -> Callable[..., BaseExecutionOutput]:
        # Use functools.wraps to preserve the metadata of the original function
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> BaseExecutionOutput:
            # Call the original function
            result = func(*args, **kwargs)
            # Check if the return type of the function is as expected
            if result == None:
                return result
            if not isinstance(result, BaseExecutionOutput):
                raise TypeError(f"Expected return type {BaseExecutionOutput.__name__}, got {type(result).__name__}")
            result.mode = "trigger"
            result.event = event_name
            return result
        # Mark the function with the event name
        wrapper._event_name = event_name
        return wrapper
    return decorator

def register_function(function_name: str) -> Callable:
    def decorator(func: Callable[..., BaseExecutionOutput]) -> Callable[..., BaseExecutionOutput]:
        """Register a given function as a handler for a YAML function named `function_name`."""
        # Use functools.wraps to preserve the metadata of the original function
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> BaseExecutionOutput:
            # Call the original function
            result = func(*args, **kwargs)
            # Check if the return type of the function is as expected
            if result == None:
                return result
            if not isinstance(result, BaseExecutionOutput):
                raise TypeError(f"Expected return type {BaseExecutionOutput.__name__}, got {type(result).__name__}")
            result.mode = "function"
            result.event = function_name
            return result
        # Mark the function with the event name
        wrapper._function_name = function_name
        return wrapper
    return decorator

class HandlerType(type):
    def __new__(cls, name, bases, namespace):
        cls_instance = super().__new__(cls, name, bases, namespace)
        cls_instance.trigger_handlers = {}
        cls_instance.functions = {}
        # Register marked methods as handlers
        for attr_name, attr_value in namespace.items():
            if callable(attr_value) and hasattr(attr_value, '_event_name'):
                event_name = getattr(attr_value, '_event_name')
                cls_instance.trigger_handlers[event_name] = attr_value
            if callable(attr_value) and hasattr(attr_value, '_function_name'):
                output_name = getattr(attr_value, '_function_name')
                cls_instance.functions[output_name] = attr_value
        return cls_instance
    
class DefaultPipelineBlock(metaclass=HandlerType):
    trigger_handlers: Dict[str, Callable]
    functions: Dict[str, Callable]
    def __init__(self, config, instance_config, credentials, memory=None):
        super().__init__()
        self.config = config
        self.instance_config = instance_config
        self.credentials=credentials
        self.memory = memory

        config_dict = config.dict()
        self.instance_config = self.validate_instance_config(self.instance_config).model_dump()
        if 'triggers' in config_dict['actions']:
            assert hasattr(self, 'trigger_handlers'), "No trigger handlers are defined"
            for trigger in config_dict['actions']['triggers']:
                if trigger['event'] not in self.trigger_handlers:
                    raise ValueError(f"Trigger event {trigger['event']} defined in manifest, but no handler is defined!")
        if 'functions' in config_dict['actions']:
            assert hasattr(self, 'functions'), "No functions are defined"
            for output in config_dict['actions']['functions']:
                if output['name'] not in self.functions:
                    raise ValueError(f"Function {output['name']} defined in manifest, but no handler is defined!")
                sig = inspect.signature(self.functions[output['name']])
                params = {
                    param.name: param.annotation.__name__ for param in sig.parameters.values() if param.name != "self"
                }
                if len(params) != len(output['parameters']):
                    raise ValueError(f"Function {output['name']} specification defines {len(output['parameters'])} parameters but code defines {len(params)} parameters!")
                for item in output['parameters']:
                    name = item['name']
                    if name not in params.keys():
                        raise ValueError(f"Function {output['name']} parameter '{name}' not defined!")
                    if item['type'] != params[name]:
                        raise ValueError(f"Function {output['name']} parameter '{name}' types do not match specification '{params[name]}'!")
                    

                    
    def validate_instance_config(self, parameters: dict) -> BaseModel:
        """Code to validate the instance configuration. This function must raise an error if the configuration is invalid."""
        raise NotImplementedError("You must implement an instance configuration validator!")

class StatefulPipelineBlock(DefaultPipelineBlock):
    def __init__(self, config, instance_config):
        super().__init__(config=config, instance_config=instance_config)
    def listener_start(self):
        raise NotImplementedError("You must implement a listener start function!")
    def listener_stop(self):
        raise NotImplementedError("You must implement a listener stop function!")
    
def load_block(full_class_string):
    """
    Dynamically imports a class from a given string, handling nested modules.

    :param full_class_string: 'package.module.ClassName'
    :return: class object
    """
    module_name, class_name = full_class_string.rsplit('.', 1)
    module = importlib.import_module(module_name)
    cls = getattr(module, class_name)
    return cls


from pydantic import BaseModel, Field, RootModel, ValidationError, validator
from typing import Any, Dict, Optional
from google.protobuf import json_format

class BaseExecutionOutput(BaseModel):
    walltime: float | None = None # populated by stage runnner
    stdout: str | None = None # populated by stage runnner
    stderr: str | None = None # populated by stage runnner
    statusCode: int # 0 = SUCCESS. provided by user, or overidden by stage runner if exception thrown
    mode: str | None = None # injected by registration handlers
    event: str | None = None # injected by registration handlers
    output: str = None # actual output

def serialize_output(output: BaseExecutionOutput):
    raise NotImplementedError()


import yaml

from deplyai._schema import Pipeline
from deplyai._utils.graph import cyclic

from pydantic import ValidationError

def read_pipeline_config(yaml_path, return_graph = False):
    """Reads the Pipeline definition file at `yaml_path` and returns the Pipeline object.
    :param yaml_path: file path to the Pipeline defintion file.
    :throws pydantic_core._pydantic_core.ValidationError if schema constraints are violated.
    :throws ValueError if uniqueness constraints are violated.
    """
    with open(yaml_path, 'r') as file:
        data = yaml.safe_load(file)
    return parse_pipeline_config(data, return_graph)

def parse_pipeline_config(data, return_graph = False):
    pipeline = Pipeline(**data)
    config = pipeline.model_dump()
    graph = {}
    for item in config['stages']:
        stage_name = item['id']
        has_triggers = 'triggers' in item and len(item['triggers']) > 0
        if has_triggers:
            for trg in item['triggers']:
                name = trg['name']
                if name in graph:
                    raise ValidationError(f"Trigger name  '{name}' was used more than once!")
                graph[name] = {}
        if len(item['outputs']) == 0:
            for key in graph:
                graph[key][item['id']] = []
        else:
            for out in item['outputs']:
                name = out['name']
                dest = out['destination']
                if name not in graph:
                    raise ValidationError(f"Output name '{name}' is not a valid trigger name!")
                if stage_name not in graph[name]:
                    graph[name][stage_name] = [dest]
                else:
                    graph[name][stage_name].append(dest)
    for subgraph in graph:
        if cyclic(graph[subgraph]):
            raise ValidationError(f"Graph for trigger {subgraph} is cyclic!")
    if return_graph:
        return pipeline, graph
    return pipeline


import ast

class SafeEval(ast.NodeVisitor):
    def __init__(self):
        self.allowed_functions = ['compare', 'regexp']
        self.allowed_operators = [ast.And, ast.Or, ast.Eq, ast.NotEq, ast.Lt, ast.LtE, ast.Gt, ast.GtE]

    def visit_Call(self, node):
        if isinstance(node.func, ast.Name) and node.func.id in self.allowed_functions:
            self.generic_visit(node)
            return node
        raise ValueError("Function not allowed")

    def visit_BinOp(self, node):
        if isinstance(node.op, self.allowed_operators):
            self.generic_visit(node)
            return node
        raise ValueError("Operator not allowed")

    def visit_Name(self, node):
        if node.id == 'data' and isinstance(node.ctx, ast.Load):
            return node
        print(node.id)
        raise ValueError("Variable not allowed")

    def visit_Import(self, node):
        raise ValueError("Import not allowed")

    def visit_ImportFrom(self, node):
        raise ValueError("Import from not allowed")

    def visit_Expr(self, node):
        if not isinstance(node.value, (ast.Call, ast.BinOp, ast.Name, ast.Subscript)):
            raise ValueError("Expression not allowed")
        self.generic_visit(node)
        return node

    def visit_Load(self, node):
        return node

    def generic_visit(self, node):
        if isinstance(node, ast.AST):
            for field in node._fields:
                if isinstance(getattr(node, field), list):
                    for item in getattr(node, field):
                        if isinstance(item, ast.AST):
                            self.visit(item)
                elif isinstance(getattr(node, field), ast.AST):
                    self.visit(getattr(node, field))

def safe_condition_eval(expr, data):
    tree = ast.parse(expr, mode='eval')
    SafeEval().visit(tree)
    try:
        return eval(compile(tree, '', 'eval'), {'data': data})
    except Exception as e:
        print(e)
        return None

if __name__ == "__main__":
    data = {'platform': 'slack', 'event_id': '1717004222.968299', 'user': 'U06UFES01RR', 'user_friendly_name': 'Dev Singh', 'team': 'T06U11UCK0F', 'channel': 'C0739CF73U6', 'channel_friendly_name': 'testing', 'text': '@DeplyAI test', 'ts': '1717004222.968299', 'attachments': None, 'thread_ts': None, 'thread_status': 1, 'edited': None, 'hidden': False, 'deleted': None, 'reactions': None}
    assert safe_condition_eval("'fds' in data['text'] and data['user_friendly_name'] in ['Dev Singh', 'Ishaan Singh']", data) == False
    assert safe_condition_eval("('Migration Project' in data['text'] or 'test' in data['text']) and data['user_friendly_name'] in ['Dev Singh', 'Ishaan Singh']", data) == True
    print("Tests passed!")


from setuptools import setup, find_packages
import os
lib_folder = os.path.dirname(os.path.realpath(__file__))
requirement_path = f"{lib_folder}/requirements.txt"
install_requires = [] # Here we'll add: ["gunicorn", "docutils>=0.3", "lxml==0.5a7"]
if os.path.isfile(requirement_path):
    with open(requirement_path) as f:
        install_requires = f.read().splitlines()

setup(name="deplyai",
    version='0.2.2',
    description="Python SDK for DeplyAI's integrated AI agent platform.",
    setup_requires=["wheel"],
    license="BSD-3-Clause",
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=install_requires
)


def generate_id(tenant, pipeline_id, version, execution_uuid, stage_id) -> int:
    # Ensure inputs are within their bit limits
    tenant &= 0xFFFFFFFF         # 32 bits
    pipeline_id &= 0xFFFF        # 16 bits
    version &= 0xFFFF            # 16 bits
    execution_uuid &= 0xFFFFFFFF  # 32 bits
    stage_id &= 0xFFFFFFFF       # 32 bits

    # Shift and combine the values to fit them in 128 bits
    result: int = (tenant << 96) | (pipeline_id << 80) | (version << 64) | (execution_uuid << 32) | stage_id

    return result

def update_id(id_128bit, tenant=None, pipeline_id=None, version=None, execution_uuid=None, stage_id=None) -> int:
    # Masks for extracting each field from the existing ID
    tenant_mask = 0xFFFFFFFF << 96
    pipeline_id_mask = 0xFFFF << 80
    version_mask = 0xFFFF << 64
    execution_uuid_mask = 0xFFFFFFFF << 32
    stage_id_mask = 0xFFFFFFFF

    # Extract current values if None is passed
    current_tenant = (id_128bit & tenant_mask) >> 96
    current_pipeline_id = (id_128bit & pipeline_id_mask) >> 80
    current_version = (id_128bit & version_mask) >> 64
    current_execution_uuid = (id_128bit & execution_uuid_mask) >> 32
    current_stage_id = id_128bit & stage_id_mask

    # Update the fields if new values are provided
    tenant = current_tenant if tenant is None else tenant
    pipeline_id = current_pipeline_id if pipeline_id is None else pipeline_id
    version = current_version if version is None else version
    execution_uuid = current_execution_uuid if execution_uuid is None else execution_uuid
    stage_id = current_stage_id if stage_id is None else stage_id

    # Combine the possibly updated values back into a single 128-bit integer
    updated_id = (tenant << 96) | (pipeline_id << 80) | (version << 64) | (execution_uuid << 32) | stage_id
    return updated_id


def parse_id(id_128bit):
    # Masks for extracting each field from the ID
    tenant_mask = 0xFFFFFFFF << 96
    pipeline_id_mask = 0xFFFF << 80
    version_mask = 0xFFFF << 64
    execution_uuid_mask = 0xFFFFFFFF << 32
    stage_id_mask = 0xFFFFFFFF

    # Extract each component by applying the mask and then right-shifting to the LSB
    tenant = (id_128bit & tenant_mask) >> 96
    pipeline_id = (id_128bit & pipeline_id_mask) >> 80
    version = (id_128bit & version_mask) >> 64
    execution_uuid = (id_128bit & execution_uuid_mask) >> 32
    stage_id = id_128bit & stage_id_mask

    # Return a dictionary with the parsed values
    return {
        'tenant': tenant,
        'pipeline_id': pipeline_id,
        'version': version,
        'execution_uuid': execution_uuid,
        'stage_id': stage_id
    }

def format_id_as_uuid(id_128bit):
    hex_str = format(id_128bit, '032x')
    uuid_str = f'{hex_str[:8]}-{hex_str[8:12]}-{hex_str[12:16]}-{hex_str[16:20]}-{hex_str[20:]}'
    return uuid_str

def uuid_to_128bit(uuid_str):
    hex_str = uuid_str.replace('-', '')
    id_128bit = int(hex_str, 16)
    return id_128bit
